#include <cmath>
#include <numbers>
#include <QMutex>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shapes/circle.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shapes/triangle.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shader.hpp>
#include "../variants.hpp"
#include "../sprite_data.hpp"

namespace BGTU::ComputerGraphicsLabWork::Impl::Variants::Lab2 {
	namespace {
		class HsvSectorShader : public Utilities::Shader {
		private:
			RendererApi::PointF<2>::component_t cx, cy;
			double angle_offset_radians;
		public:
			HsvSectorShader(RendererApi::PointF<2>::component_t cx, RendererApi::PointF<2>::component_t cy, double angle_offset_radians) : cx{cx}, cy{cy}, angle_offset_radians{angle_offset_radians} {}

			HsvSectorShader(RendererApi::PointF<2> c, double angle_offset_radians) : cx{c.x}, cy{c.y}, angle_offset_radians{angle_offset_radians} {}

			HsvSectorShader(RendererApi::PointI<2> c, double angle_offset_radians) : cx{(double) c.x}, cy{(double) c.y}, angle_offset_radians{angle_offset_radians} {}

			[[nodiscard]] RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const final {
				double angle = std::atan2(static_cast<double>(y - this->cy), static_cast<double>(x - this->cx)) + this->angle_offset_radians;
				return Utilities::hsv_to_rgb_360_1_1(std::fmod(std::fmod(angle / std::numbers::pi_v<double> * 180, 360.0) + 360, 360.0), 1.0, 1.0);
			}
		};

		template<double z, RendererApi::Color::Transparent edge_color, double center_distance_multiplier, double center_angle_degrees, double radius_multiplier>
		class Circle : public RendererApi::Sprite<SpriteData::ShapeData, Utilities::ZoomedVoxelPainter<Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl>> {
		private:
			template<RendererApi::Color::Transparent c1, RendererApi::Color::Transparent c2>
			class RadialShader : public Utilities::Shader {
			private:
				RendererApi::PointF<2>::component_t cx, cy;
				RendererApi::PointF<2>::component_t radius;
			public:
				RadialShader(RendererApi::PointF<2>::component_t cx, RendererApi::PointF<2>::component_t cy, RendererApi::PointF<2>::component_t radius) : cx{cx}, cy{cy}, radius{radius} {}

				RadialShader(RendererApi::PointF<2> c, RendererApi::PointF<2>::component_t radius) : cx{c.x}, cy{c.y}, radius{radius} {}

				RadialShader(RendererApi::PointI<2> c, RendererApi::PointF<2>::component_t radius) : cx{(double) c.x}, cy{(double) c.y}, radius{radius} {}

				[[nodiscard]] RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const final {
					RendererApi::PointF<2>::component_t distance = std::hypot(x - this->cx, y - this->cy);
					double ratio = distance / this->radius;
					float c1v;

					if (ratio < 0.5) {
						c1v = (float) (std::fabs(ratio - 0.25) * 4);
					} else if (ratio < 1.0) {
						c1v = (float) (std::fabs(ratio - 0.75) * 4);
					} else {
						return c1;
					}

					return Utilities::colors_gradient(c1, c1v, c2);
				}
			};

			enum class ShaderType : unsigned {
				SOLID_RED,
				SOLID_GREEN,
				SOLID_BLUE,
				RADIAL_DARK_RED,
				RADIAL_DARK_GREEN,
				RADIAL_DARK_BLUE,
				RADIAL_LIGHT_RED,
				RADIAL_LIGHT_GREEN,
				RADIAL_LIGHT_BLUE,
				RADIAL_TRANSPARENT_RED,
				RADIAL_TRANSPARENT_GREEN,
				RADIAL_TRANSPARENT_BLUE,
				SECTOR_STATIC,
				SECTOR_ROTATED_POS,
				SECTOR_ROTATED_NEG,
			};

			static inline ShaderType const next_shader_type[] = {
					ShaderType::SOLID_GREEN,
					ShaderType::SOLID_BLUE,
					ShaderType::RADIAL_DARK_RED,
					ShaderType::RADIAL_DARK_GREEN,
					ShaderType::RADIAL_DARK_BLUE,
					ShaderType::RADIAL_LIGHT_RED,
					ShaderType::RADIAL_LIGHT_GREEN,
					ShaderType::RADIAL_LIGHT_BLUE,
					ShaderType::RADIAL_TRANSPARENT_RED,
					ShaderType::RADIAL_TRANSPARENT_GREEN,
					ShaderType::RADIAL_TRANSPARENT_BLUE,
					ShaderType::SECTOR_STATIC,
					ShaderType::SECTOR_ROTATED_POS,
					ShaderType::SECTOR_ROTATED_NEG,
					ShaderType::SOLID_RED
			};

			QMutex sync;
			ShaderType current_shader;

			RendererApi::PointI<2> center(SpriteData::ShapeData const *data) const {
				return data->pos_rotated(data->radius * center_distance_multiplier, center_angle_degrees);
			}

			template<class receiver_t>
			void _select_shader(SpriteData::ShapeData const *data, receiver_t receiver) const {
				const_cast<QMutex *>(&(this->sync))->lock();
				ShaderType t = this->current_shader;
				const_cast<QMutex *>(&(this->sync))->unlock();
				switch (t) {
					case ShaderType::SOLID_RED: {
						auto s = Utilities::MonoShader::static_<RendererApi::Color{255, 0, 0}>;
						receiver(s);
						return;
					}
					case ShaderType::SOLID_GREEN: {
						auto s = Utilities::MonoShader::static_<RendererApi::Color{0, 255, 0}>;
						receiver(s);
						return;
					}
					case ShaderType::SOLID_BLUE: {
						auto s = Utilities::MonoShader::static_<RendererApi::Color{0, 0, 255}>;
						receiver(s);
						return;
					}
					case ShaderType::RADIAL_DARK_RED: {
						RadialShader<RendererApi::Color{255, 0, 0}, RendererApi::Color{0, 0, 0}> s{this->center(data), data->radius * radius_multiplier - 1};
						receiver(s);
						return;
					}
					case ShaderType::RADIAL_DARK_GREEN: {
						RadialShader<RendererApi::Color{0, 255, 0}, RendererApi::Color{0, 0, 0}> s{this->center(data), data->radius * radius_multiplier - 1};
						receiver(s);
						return;
					}
					case ShaderType::RADIAL_DARK_BLUE: {
						RadialShader<RendererApi::Color{0, 0, 255}, RendererApi::Color{0, 0, 0}> s{this->center(data), data->radius * radius_multiplier - 1};
						receiver(s);
						return;
					}
					case ShaderType::RADIAL_LIGHT_RED: {
						RadialShader<RendererApi::Color{255, 0, 0}, RendererApi::Color{255, 255, 255}> s{this->center(data), data->radius * radius_multiplier - 1};
						receiver(s);
						return;
					}
					case ShaderType::RADIAL_LIGHT_GREEN: {
						RadialShader<RendererApi::Color{0, 255, 0}, RendererApi::Color{255, 255, 255}> s{this->center(data), data->radius * radius_multiplier - 1};
						receiver(s);
						return;
					}
					case ShaderType::RADIAL_LIGHT_BLUE: {
						RadialShader<RendererApi::Color{0, 0, 255}, RendererApi::Color{255, 255, 255}> s{this->center(data), data->radius * radius_multiplier - 1};
						receiver(s);
						return;
					}
					case ShaderType::RADIAL_TRANSPARENT_RED: {
						RadialShader<RendererApi::Color::Transparent{255, 0, 0, 255}, RendererApi::Color::Transparent{255, 0, 0, 0}> s{this->center(data), data->radius * radius_multiplier - 1};
						receiver(s);
						return;
					}
					case ShaderType::RADIAL_TRANSPARENT_GREEN: {
						RadialShader<RendererApi::Color::Transparent{0, 255, 0, 255}, RendererApi::Color::Transparent{0, 255, 0, 0}> s{this->center(data), data->radius * radius_multiplier - 1};
						receiver(s);
						return;
					}
					case ShaderType::RADIAL_TRANSPARENT_BLUE: {
						RadialShader<RendererApi::Color::Transparent{0, 0, 255, 255}, RendererApi::Color::Transparent{0, 0, 255, 0}> s{this->center(data), data->radius * radius_multiplier - 1};
						receiver(s);
						return;
					}
					case ShaderType::SECTOR_STATIC: {
						HsvSectorShader s{this->center(data), 0};
						receiver(s);
						return;
					}
					case ShaderType::SECTOR_ROTATED_POS: {
						HsvSectorShader s{this->center(data), data->rotation_radians};
						receiver(s);
						return;
					}
					case ShaderType::SECTOR_ROTATED_NEG: {
						HsvSectorShader s{this->center(data), -data->rotation_radians};
						receiver(s);
						return;
					}
				}
			};
		public:
			Circle() : sync{}, current_shader{ShaderType::SOLID_RED} {}


			Circle(Circle &&other) noexcept: sync{}, current_shader{other.current_shader} {

			}

			void draw(Utilities::ZoomedVoxelPainter<Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame, SpriteData::ShapeData const *data) const final {
				this->_select_shader(
						data,
						[&]<class shader_t>(shader_t const &shader) {
							Utilities::Shapes::iterate_circle_fill(
									this->center(data),
									data->radius * radius_multiplier,
									[&](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) {
										if (is_edge) {
											frame->add_voxel(x, y, z, edge_color);
										} else {
											frame->add_voxel(x, y, z, shader.get_color(x, y));
										}
									}
							);
						}
				);
			}

			void clicked() final {
				this->sync.lock();
				this->current_shader = next_shader_type[(unsigned) this->current_shader];
				this->sync.unlock();
			}
		};

		template<double z, RendererApi::Color::Transparent edge_color, double radius_multiplier>
		class Triangle : public RendererApi::Sprite<SpriteData::ShapeData, Utilities::ZoomedVoxelPainter<Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl>> {
		private:
			enum class ShaderType : unsigned {
				SOLID_RED,
				SOLID_GREEN,
				SOLID_BLUE,
				SECTOR_STATIC,
				SECTOR_ROTATED_POS,
				SECTOR_ROTATED_NEG,
				BARYCENTRIC,
				BARYCENTRIC_SEMI_TRANSPARENT,
			};

			static inline ShaderType const next_shader_type[] = {
					ShaderType::SOLID_GREEN,
					ShaderType::SOLID_BLUE,
					ShaderType::SECTOR_STATIC,
					ShaderType::SECTOR_ROTATED_POS,
					ShaderType::SECTOR_ROTATED_NEG,
					ShaderType::BARYCENTRIC,
					ShaderType::BARYCENTRIC_SEMI_TRANSPARENT,
					ShaderType::SOLID_RED,
			};

			QMutex sync;
			ShaderType current_shader;

			template<class receiver_t>
			void _select_shader(SpriteData::ShapeData const *data, receiver_t receiver) const {
				const_cast<QMutex *>(&(this->sync))->lock();
				ShaderType t = this->current_shader;
				const_cast<QMutex *>(&(this->sync))->unlock();
				switch (t) {
					case ShaderType::SOLID_RED: {
						auto s = Utilities::MonoShader::static_<RendererApi::Color{255, 0, 0}>;
						receiver(s);
						return;
					}
					case ShaderType::SOLID_GREEN: {
						auto s = Utilities::MonoShader::static_<RendererApi::Color{0, 255, 0}>;
						receiver(s);
						return;
					}
					case ShaderType::SOLID_BLUE: {
						auto s = Utilities::MonoShader::static_<RendererApi::Color{0, 0, 255}>;
						receiver(s);
						return;
					}
					case ShaderType::SECTOR_STATIC: {
						HsvSectorShader s{0, 0, 0};
						receiver(s);
						return;
					}
					case ShaderType::SECTOR_ROTATED_POS: {
						HsvSectorShader s{0, 0, data->rotation_radians};
						receiver(s);
						return;
					}
					case ShaderType::SECTOR_ROTATED_NEG: {
						HsvSectorShader s{0, 0, -data->rotation_radians};
						receiver(s);
						return;
					}
					case ShaderType::BARYCENTRIC: {
						Utilities::Shapes::triangle_barycentric_shader s{
								data->pos_rotated_f(data->radius * radius_multiplier, -120.0), RendererApi::Color{255, 0, 0},
								data->pos_rotated_f(data->radius * radius_multiplier, 0.0), RendererApi::Color{0, 255, 0},
								data->pos_rotated_f(data->radius * radius_multiplier, 120.0), RendererApi::Color{0, 0, 255}
						};
						receiver(s);
						return;
					}
					case ShaderType::BARYCENTRIC_SEMI_TRANSPARENT: {
						Utilities::Shapes::triangle_barycentric_shader s{
								data->pos_rotated_f(data->radius * radius_multiplier, -120.0), RendererApi::Color::Transparent{255, 0, 0, 127},
								data->pos_rotated_f(data->radius * radius_multiplier, 0.0), RendererApi::Color::Transparent{0, 255, 0, 127},
								data->pos_rotated_f(data->radius * radius_multiplier, 120.0), RendererApi::Color::Transparent{0, 0, 255, 127}
						};
						receiver(s);
						return;
					}
				}
			};
		public:
			Triangle() : sync{}, current_shader{ShaderType::SOLID_RED} {}


			Triangle(Triangle &&other) noexcept: sync{}, current_shader{other.current_shader} {}

			void draw(Utilities::ZoomedVoxelPainter<Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl> *frame, SpriteData::ShapeData const *data) const final {
				this->_select_shader(
						data,
						[&]<class shader_t>(shader_t const &shader) {
							Utilities::Shapes::iterate_triangle_fill(
									data->pos_rotated(data->radius * radius_multiplier, -120.0),
									data->pos_rotated(data->radius * radius_multiplier, 0.0),
									data->pos_rotated(data->radius * radius_multiplier, 120.0),
									[&](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) {
										if (is_edge) {
											frame->add_voxel(x, y, z, edge_color);
										} else {
											frame->add_voxel(x, y, z, shader.get_color(x, y));
										}
									}
							);
						}
				);
			}

			void clicked() final {
				this->sync.lock();
				this->current_shader = next_shader_type[(unsigned) this->current_shader];
				this->sync.unlock();
			}
		};
	}
	variant_sprites variant3 = variant_sprites::make(
			Circle<4.0, {255, 255, 255}, 0.0, 0.0, 1.0>{},
			Triangle<3.0, {255, 255, 255}, 1.0>{},
			Circle<2.0, {255, 255, 255}, 0.6666666666666, 0.0, 0.16666666666666666>{},
			Circle<2.0, {255, 255, 255}, 0.6666666666666, 120.0, 0.16666666666666666>{},
			Circle<2.0, {255, 255, 255}, 0.6666666666666, -120.0, 0.16666666666666666>{},
			Circle<1.0, {255, 255, 255}, 0.0, 0.0, 0.5>{}
	);
}