#pragma once

#include <cstdlib>
#include <cmath>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>

namespace BGTU::ComputerGraphicsLabWork::Utilities {

	template<class elem_t>
	class Matrix3 {
	private:
		elem_t data[3][3];

		inline Matrix3() = default;

	public:

		inline Matrix3(
				elem_t e00, elem_t e01, elem_t e02,
				elem_t e10, elem_t e11, elem_t e12,
				elem_t e20, elem_t e21, elem_t e22
		) : data{e00, e01, e02, e10, e11, e12, e20, e21, e22} {}

		static const Matrix3<elem_t> I;
	private:
		static const Matrix3<elem_t> _tmp;


		template<class operator_t>
		Matrix3<elem_t> _zipmap_matrix(Matrix3<elem_t> const *other, operator_t op) const {
			Matrix3<elem_t> dst;
			auto _map_row = [&](std::size_t row_index) {
				dst.data[row_index][0] = op(row_index, 0, this->data[row_index][0], other->data[row_index][0]);
				dst.data[row_index][1] = op(row_index, 1, this->data[row_index][1], other->data[row_index][1]);
				dst.data[row_index][2] = op(row_index, 2, this->data[row_index][2], other->data[row_index][2]);
			};
			_map_row(0);
			_map_row(1);
			_map_row(2);

			return dst;
		}

	public:
		inline Matrix3<elem_t> operator+(Matrix3<elem_t> const &other) const {
			return this->_zipmap_matrix(&other, [](auto, auto, auto s, auto o) { return s + o; });
		}

		inline Matrix3<elem_t> operator-(Matrix3<elem_t> const &other) const {
			return this->_zipmap_matrix(&other, [](auto, auto, auto s, auto o) { return s - o; });
		}

		inline Matrix3<elem_t> operator*(int other) const {
			return this->_zipmap_matrix(&Matrix3::_tmp, [=](auto, auto, auto s, auto) { return s * other; });
		}

		inline Matrix3<elem_t> operator*(long other) const {
			return this->_zipmap_matrix(&Matrix3::_tmp, [=](auto, auto, auto s, auto) { return s * other; });
		}

		inline Matrix3<elem_t> operator*(float other) const {
			return this->_zipmap_matrix(&Matrix3::_tmp, [=](auto, auto, auto s, auto) { return s * other; });
		}

		inline Matrix3<elem_t> operator*(double other) const {
			return this->_zipmap_matrix(&Matrix3::_tmp, [=](auto, auto, auto s, auto) { return s * other; });
		}

		inline Matrix3<elem_t> operator/(int other) const {
			return this->_zipmap_matrix(&Matrix3::_tmp, [=](auto, auto, auto s, auto) { return s / other; });
		}

		inline Matrix3<elem_t> operator/(long other) const {
			return this->_zipmap_matrix(&Matrix3::_tmp, [=](auto, auto, auto s, auto) { return s / other; });
		}

		inline Matrix3<elem_t> operator/(float other) const {
			return this->_zipmap_matrix(&Matrix3::_tmp, [=](auto, auto, auto s, auto) { return s / other; });
		}

		inline Matrix3<elem_t> operator/(double other) const {
			return this->_zipmap_matrix(&Matrix3::_tmp, [=](auto, auto, auto s, auto) { return s / other; });
		}

		inline Matrix3<elem_t> operator*(Matrix3<elem_t> const &other) const {
			return this->_zipmap_matrix(&Matrix3::_tmp, [&](auto y, auto x, auto, auto) {
				return this->data[y][0] * other.data[0][x] + this->data[y][1] * other.data[1][x] + this->data[y][2] * other.data[2][x];
			});
		}

		static Matrix3<elem_t> scale(elem_t x, elem_t y) {
			return Matrix3<elem_t>{x, 0, 0,
								   0, y, 0,
								   0, 0, 1};
		}

		static Matrix3<elem_t> scale(elem_t s) {
			return Matrix3<elem_t>{s, 0, 0,
								   0, s, 0,
								   0, 0, 1};
		}

		static Matrix3<elem_t> rotate(elem_t radians) {
			auto s = std::sin(radians);
			auto c = std::cos(radians);
			return Matrix3<elem_t>{c, -s, 0,
								   s, c, 0,
								   0, 0, 1};
		}

		static Matrix3<elem_t> shift(elem_t x, elem_t y) {
			return Matrix3<elem_t>{1, 0, x,
								   0, 1, y,
								   0, 0, 1};
		}

		RendererApi::Point<2, elem_t> operator*(RendererApi::Point<2, elem_t> other) const {
			return RendererApi::Point<2, elem_t>{
					this->data[0][0] * other.x + this->data[0][1] * other.y + this->data[0][2],
					this->data[1][0] * other.x + this->data[1][1] * other.y + this->data[1][2],
			};
		}

		elem_t det() const {
			return this->data[0][0] * (this->data[1][1] * this->data[2][2] - this->data[1][2] * this->data[2][1]) -
				   this->data[0][1] * (this->data[1][0] * this->data[2][2] - this->data[1][2] * this->data[2][0]) +
				   this->data[0][2] * (this->data[1][0] * this->data[2][1] - this->data[1][1] * this->data[2][0]);
		}

		Matrix3<elem_t> adj() const {
			return Matrix3<elem_t>{
					this->data[1][1] * this->data[2][2] - this->data[1][2] * this->data[2][1],
					-(this->data[1][0] * this->data[2][2] - this->data[1][2] * this->data[2][0]),
					this->data[1][0] * this->data[2][1] - this->data[1][1] * this->data[2][0],
					-(this->data[0][1] * this->data[2][2] - this->data[0][2] * this->data[2][1]),
					this->data[0][0] * this->data[2][2] - this->data[0][2] * this->data[2][0],
					-(this->data[0][0] * this->data[2][1] - this->data[0][1] * this->data[2][0]),
					this->data[0][1] * this->data[1][2] - this->data[0][2] * this->data[1][1],
					-(this->data[0][0] * this->data[1][2] - this->data[0][2] * this->data[1][0]),
					this->data[0][0] * this->data[1][1] - this->data[0][1] * this->data[1][0]
			};
		}

		Matrix3<elem_t> transposed() const {
			return Matrix3<elem_t>{
					this->data[0][0], this->data[1][0], this->data[2][0],
					this->data[0][1], this->data[1][1], this->data[2][1],
					this->data[0][2], this->data[1][2], this->data[2][2]
			};
		}

		Matrix3<elem_t> inversed() const {
			return this->adj().transposed() / this->det();
		}
	};

	using Matrix3i = Matrix3<int>;
	using Matrix3l = Matrix3<long>;
	using Matrix3f = Matrix3<float>;
	using Matrix3d = Matrix3<double>;

	template<class elem_t>
	class Matrix4;

	template<class elem_t>
	class Vec4 {
	private:
		elem_t data[4];


		template<class>
		friend
		class Matrix4;

	public:
		Vec4(elem_t a, elem_t b, elem_t c, elem_t d) : data{a, b, c, d} {}

		elem_t operator[](std::size_t i) const {
			return this->data[i];
		};
	};

	template<class elem_t>
	class Matrix4 {
	private:
		elem_t data[4][4];

		Matrix4() = default;

	public:
#if 0
		inline Matrix4(
			std::initializer_list<elem_t> r0,
			std::initializer_list<elem_t> r1,
			std::initializer_list<elem_t> r2,
			std::initializer_list<elem_t> r3
	) : data{r0, r1, r2, r3} {}
#endif

		inline Matrix4(
				elem_t e00, elem_t e01, elem_t e02, elem_t e03,
				elem_t e10, elem_t e11, elem_t e12, elem_t e13,
				elem_t e20, elem_t e21, elem_t e22, elem_t e23,
				elem_t e30, elem_t e31, elem_t e32, elem_t e33
		) : data{e00, e01, e02, e03, e10, e11, e12, e13, e20, e21, e22, e23, e30, e31, e32, e33} {}

		static const Matrix4<elem_t> I;

	private:
		static const Matrix4<elem_t> _tmp;

		template<class operator_t>
		Matrix4<elem_t> _zipmap_matrix(Matrix4<elem_t> const *other, operator_t op) const {
			Matrix4<elem_t> dst;
			auto _map_row = [&](std::size_t row_index) {
				dst.data[row_index][0] = op(row_index, 0, this->data[row_index][0], other->data[row_index][0]);
				dst.data[row_index][1] = op(row_index, 1, this->data[row_index][1], other->data[row_index][1]);
				dst.data[row_index][2] = op(row_index, 2, this->data[row_index][2], other->data[row_index][2]);
				dst.data[row_index][3] = op(row_index, 3, this->data[row_index][3], other->data[row_index][3]);
			};
			_map_row(0);
			_map_row(1);
			_map_row(2);
			_map_row(3);

			return dst;
		}

	public:
		inline Matrix4<elem_t> operator+(Matrix4<elem_t> const &other) const {
			return this->_zipmap_matrix(&other, [](auto, auto, auto s, auto o) { return s + o; });
		}

		inline Matrix4<elem_t> operator-(Matrix4<elem_t> const &other) const {
			return this->_zipmap_matrix(&other, [](auto, auto, auto s, auto o) { return s - o; });
		}

		inline Matrix4<elem_t> operator*(int other) const {
			return this->_zipmap_matrix(&Matrix4::_tmp, [=](auto, auto, auto s, auto) { return s * other; });
		}

		inline Matrix4<elem_t> operator*(long other) const {
			return this->_zipmap_matrix(&Matrix4::_tmp, [=](auto, auto, auto s, auto) { return s * other; });
		}

		inline Matrix4<elem_t> operator*(float other) const {
			return this->_zipmap_matrix(&Matrix4::_tmp, [=](auto, auto, auto s, auto) { return s * other; });
		}

		inline Matrix4<elem_t> operator*(double other) const {
			return this->_zipmap_matrix(&Matrix4::_tmp, [=](auto, auto, auto s, auto) { return s * other; });
		}

		inline Matrix4<elem_t> operator/(int other) const {
			return this->_zipmap_matrix(&Matrix4::_tmp, [=](auto, auto, auto s, auto) { return s / other; });
		}

		inline Matrix4<elem_t> operator/(long other) const {
			return this->_zipmap_matrix(&Matrix4::_tmp, [=](auto, auto, auto s, auto) { return s / other; });
		}

		inline Matrix4<elem_t> operator/(float other) const {
			return this->_zipmap_matrix(&Matrix4::_tmp, [=](auto, auto, auto s, auto) { return s / other; });
		}

		inline Matrix4<elem_t> operator/(double other) const {
			return this->_zipmap_matrix(&Matrix4::_tmp, [=](auto, auto, auto s, auto) { return s / other; });
		}

		inline Matrix4<elem_t> operator*(Matrix4<elem_t> const &other) const {
			return this->_zipmap_matrix(&Matrix4::_tmp, [&](auto y, auto x, auto, auto) {
				return this->data[y][0] * other.data[0][x] + this->data[y][1] * other.data[1][x] + this->data[y][2] * other.data[2][x] + this->data[y][3] * other.data[3][x];
			});
		}

		static Matrix4<elem_t> scale(elem_t x, elem_t y, elem_t z) {
			return Matrix4<elem_t>{x, 0, 0, 0,
								   0, y, 0, 0,
								   0, 0, z, 0,
								   0, 0, 0, 1};
		}

		static Matrix4<elem_t> scale(elem_t s) {
			return Matrix4<elem_t>{s, 0, 0, 0,
								   0, s, 0, 0,
								   0, 0, s, 0,
								   0, 0, 0, 1};
		}


		static Matrix4<elem_t> rotate_x(elem_t radians) {
			auto s = std::sin(radians);
			auto c = std::cos(radians);
			return Matrix4<elem_t>{1, 0, 0, 0,
								   0, c, -s, 0,
								   0, s, c, 0,
								   0, 0, 0, 1};
		}

		static Matrix4<elem_t> rotate_y(elem_t radians) {
			auto s = std::sin(radians);
			auto c = std::cos(radians);
			return Matrix4<elem_t>{c, 0, s, 0,
								   0, 1, 0, 0,
								   -s, 0, c, 0,
								   0, 0, 0, 1};
		}

		static Matrix4<elem_t> rotate_z(elem_t radians) {
			auto s = std::sin(radians);
			auto c = std::cos(radians);
			return Matrix4<elem_t>{c, -s, 0, 0,
								   s, c, 0, 0,
								   0, 0, 1, 0,
								   0, 0, 0, 1};
		}

		static Matrix4<elem_t> shift(elem_t x, elem_t y, elem_t z) {
			return Matrix4<elem_t>{1, 0, 0, x,
								   0, 1, 0, y,
								   0, 0, 1, z,
								   0, 0, 0, 1};
		}

		RendererApi::Point<3, elem_t> operator*(RendererApi::Point<3, elem_t> other) const {
			return RendererApi::Point<3, elem_t>{
					this->data[0][0] * other.x + this->data[0][1] * other.y + this->data[0][2] * other.z + this->data[0][3],
					this->data[1][0] * other.x + this->data[1][1] * other.y + this->data[1][2] * other.z + this->data[1][3],
					this->data[2][0] * other.x + this->data[2][1] * other.y + this->data[2][2] * other.z + this->data[2][3],
			};
		}

		Vec4<elem_t> operator*(Vec4<elem_t> other) const {
			return Vec4<elem_t>{
					this->data[0][0] * other.data[0] + this->data[0][1] * other.data[1] + this->data[0][2] * other.data[2] + this->data[0][3] * other.data[3],
					this->data[1][0] * other.data[0] + this->data[1][1] * other.data[1] + this->data[1][2] * other.data[2] + this->data[1][3] * other.data[3],
					this->data[2][0] * other.data[0] + this->data[2][1] * other.data[1] + this->data[2][2] * other.data[2] + this->data[2][3] * other.data[3],
					this->data[3][0] * other.data[0] + this->data[3][1] * other.data[1] + this->data[3][2] * other.data[2] + this->data[3][3] * other.data[3]
			};
		}

		Matrix4<elem_t> transposed() const {
			return Matrix4<elem_t>{
					this->data[0][0], this->data[1][0], this->data[2][0], this->data[3][0],
					this->data[0][1], this->data[1][1], this->data[2][1], this->data[3][1],
					this->data[0][2], this->data[1][2], this->data[2][2], this->data[3][2],
					this->data[0][3], this->data[1][3], this->data[2][3], this->data[3][3],
			};
		}

		static Matrix4<elem_t> projection_orto(elem_t l, elem_t r, elem_t b, elem_t t, elem_t n, elem_t f) {
			return Matrix4<elem_t>{static_cast<elem_t>(2.0 / (r - l)), 0, 0, static_cast<elem_t>(-(r + l) / (r - l)),
								   0, static_cast<elem_t>(2.0 / (t - b)), 0, static_cast<elem_t>(-(t + b) / (t - b)),
								   0, 0, static_cast<elem_t>(2.0 / (f - n)), static_cast<elem_t>(-(f + n) / (f - n)),
								   0, 0, 0, 1};
		}

		static Matrix4<elem_t> projection_central(elem_t l, elem_t r, elem_t b, elem_t t, elem_t n, elem_t f) {
			return Matrix4<elem_t>{static_cast<elem_t>(2.0 * n / (r - l)), 0, static_cast<elem_t>((r + l) / (r - l)), 0,
								   0, static_cast<elem_t>(2.0 * n / (t - b)), static_cast<elem_t>((t + b) / (t - b)), 0,
								   0, 0, static_cast<elem_t>(-(f + n) / (f - n)), static_cast<elem_t>(-2.0 * n * f / (f - n)),
								   0, 0, -1, 0};
		}
	};

	template<class elem_t>
	const Matrix3<elem_t> Matrix3<elem_t>::I{1, 0, 0,
											 0, 1, 0,
											 0, 0, 1};

	template<class elem_t>
	const Matrix3<elem_t> Matrix3<elem_t>::_tmp{};

	template<class elem_t>
	const Matrix4<elem_t> Matrix4<elem_t>::I{1, 0, 0, 0,
											 0, 1, 0, 0,
											 0, 0, 1, 0,
											 0, 0, 0, 1};

	template<class elem_t>
	const Matrix4<elem_t> Matrix4<elem_t>::_tmp{};

	using Matrix4i = Matrix4<int>;
	using Matrix4l = Matrix4<long>;
	using Matrix4f = Matrix4<float>;
	using Matrix4d = Matrix4<double>;
}