#pragma once

#include <cmath>
#include <numbers>
#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include "color.hpp"
#include "matrix.hpp"


namespace BGTU::ComputerGraphicsLabWork::Utilities {
	class Shader {
	public:
		[[nodiscard]] virtual RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const = 0;

		[[nodiscard]] inline RendererApi::Color::Transparent get_color(RendererApi::PointF<2> p) const {
			return this->get_color(p.x, p.y);
		};
#if 0
		[[nodiscard]] inline RendererApi::Color::Transparent get_color(RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) {
			return this->get_color(static_cast<RendererApi::PointF<2>::component_t>(x), static_cast<RendererApi::PointF<2>::component_t>(y));
		};
#endif
		[[nodiscard]]  inline RendererApi::Color::Transparent get_color(RendererApi::PointI<2> p) const {
			return this->get_color((RendererApi::PointF<2>) p);
		};

		class Empty;

		static Shader::Empty const empty;
	};

	class Shader::Empty : public Shader {
	public:
		constexpr Empty() noexcept = default;

		[[nodiscard]] constexpr RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const noexcept final {
			return RendererApi::Color::Transparent{0, 0, 0, 0};
		};

		operator Shader::Empty const *() const noexcept { // NOLINT(*-explicit-constructor)
			return this;
		}

		Shader::Empty const *operator->() const noexcept {
			return this;
		}

	};


	constexpr bool operator==(Shader::Empty const &, Shader::Empty const &) noexcept {
		return true;
	}

	constexpr bool operator==(Shader::Empty const &, Shader::Empty const *) noexcept {
		return true;
	}

	constexpr bool operator==(Shader::Empty const &, Shader const *other) noexcept {
		return dynamic_cast<Shader::Empty const *>(other) != nullptr;
	}

	constexpr bool operator!=(Shader::Empty const &, Shader::Empty const &) noexcept {
		return false;
	}

	constexpr bool operator!=(Shader::Empty const &, Shader::Empty const *) noexcept {
		return false;
	}

	constexpr bool operator!=(Shader::Empty const &, Shader const *other) noexcept {
		return dynamic_cast<Shader::Empty const *>(other) == nullptr;
	}

	constexpr bool operator==(Shader::Empty const *, Shader::Empty const &) noexcept {
		return true;
	}

	constexpr bool operator==(Shader const *other, Shader::Empty const &) noexcept {
		return dynamic_cast<Shader::Empty const *>(other) != nullptr;
	}

	constexpr bool operator!=(Shader::Empty const *, Shader::Empty const &) noexcept {
		return false;
	}

	constexpr bool operator!=(Shader const *other, Shader::Empty const &) noexcept {
		return dynamic_cast<Shader::Empty const *>(other) == nullptr;
	}

	constexpr inline Shader::Empty const Shader::empty{};

	class MonoShader : public Shader {
	private:
		RendererApi::Color::Transparent c;

	public:
		constexpr explicit MonoShader(RendererApi::Color::Transparent c) noexcept: c{c} {}

		[[nodiscard]] constexpr RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const noexcept final {
			return this->c;
		};

		template<RendererApi::Color::Transparent _color>
		class Static : public Shader {
		public:
			constexpr Static() noexcept = default;

			static constexpr RendererApi::Color::Transparent const COLOR = _color;

			[[nodiscard]] constexpr RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const noexcept final {
				return _color;
			};

			operator MonoShader::Static<_color> const *() const noexcept { // NOLINT(*-explicit-constructor)
				return this;
			}


			MonoShader::Static<_color> const *operator->() const noexcept {
				return this;
			}


			template<RendererApi::Color::Transparent other_color>
			constexpr bool operator==(MonoShader::Static<other_color>) const noexcept {
				return _color == other_color;
			}
		};

		template<RendererApi::Color::Transparent color>
		static MonoShader::Static<color> const static_;
	};

	template<RendererApi::Color::Transparent color>
	constexpr MonoShader::Static<color> const MonoShader::static_{};

	template<RendererApi::Color::Transparent c1, RendererApi::Color::Transparent c2>
	class RadialShader : public Utilities::Shader {
	private:
		RendererApi::PointF<2>::component_t cx, cy;
		RendererApi::PointF<2>::component_t radius;
	public:
		RadialShader(RendererApi::PointF<2>::component_t cx, RendererApi::PointF<2>::component_t cy, RendererApi::PointF<2>::component_t radius) : cx{cx}, cy{cy}, radius{radius} {}

		RadialShader(RendererApi::PointF<2> c, RendererApi::PointF<2>::component_t radius) : cx{c.x}, cy{c.y}, radius{radius} {}

		RadialShader(RendererApi::PointI<2> c, RendererApi::PointF<2>::component_t radius) : cx{(double) c.x}, cy{(double) c.y}, radius{radius} {}

		[[nodiscard]] RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const final {
			RendererApi::PointF<2>::component_t distance = std::hypot(x - this->cx, y - this->cy);
			double ratio = distance / this->radius;
			float c1v;

			if (ratio < 0.5) {
				c1v = (float) (std::fabs(ratio - 0.25) * 4);
			} else if (ratio < 1.0) {
				c1v = (float) (std::fabs(ratio - 0.75) * 4);
			} else {
				return c1;
			}

			return Utilities::colors_gradient(c1, c1v, c2);
		}
	};

	class hsv_sector_shader : public Utilities::Shader {
	private:
		RendererApi::PointF<2>::component_t cx, cy;
		double angle_offset_radians;
	public:
		hsv_sector_shader(RendererApi::PointF<2>::component_t cx, RendererApi::PointF<2>::component_t cy, double angle_offset_radians) : cx{cx}, cy{cy}, angle_offset_radians{angle_offset_radians} {}

		hsv_sector_shader(RendererApi::PointF<2> c, double angle_offset_radians) : cx{c.x}, cy{c.y}, angle_offset_radians{angle_offset_radians} {}

		hsv_sector_shader(RendererApi::PointI<2> c, double angle_offset_radians) : cx{(double) c.x}, cy{(double) c.y}, angle_offset_radians{angle_offset_radians} {}

		[[nodiscard]] RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const final {
			double angle = std::atan2(static_cast<double>(y - this->cy), static_cast<double>(x - this->cx)) + this->angle_offset_radians;
			return Utilities::hsv_to_rgb_360_1_1(std::fmod(std::fmod(angle / std::numbers::pi_v<double> * 180, 360.0) + 360, 360.0), 1.0, 1.0);
		}
	};

	class TransformedShader : public Shader {
	private:
		Shader const *shader;
		Matrix4d transform;

	public:
		inline TransformedShader(Shader const *s, Matrix4d t) : shader{s}, transform{t} {}

		[[nodiscard]] inline RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const final {
			RendererApi::PointF<3> in{x, y, 0};
			RendererApi::PointF<3> out = this->transform * in;
			return this->shader->get_color(out.x, out.y);
		}
	};
}